@@ -8,7 +8,7 @@
 
 namespace netgen::cg
 {
-  typedef ngcore::ClosedHashTable<ngcore::INT<3,size_t>, size_t> PointTable;
+  typedef ngcore::ClosedHashTable<ngcore::IVec<3,size_t>, size_t> PointTable;
 
   int getDim(ElementType_t type)
     {
@@ -416,7 +416,7 @@ namespace netgen::cg
 
           for(auto i : Range(nv))
           {
-            ngcore::INT<3,size_t> hash = {*reinterpret_cast<size_t*>(&x[i]), *reinterpret_cast<size_t*>(&y[i]), *reinterpret_cast<size_t*>(&z[i])};
+            ngcore::IVec<3,size_t> hash = {*reinterpret_cast<size_t*>(&x[i]), *reinterpret_cast<size_t*>(&y[i]), *reinterpret_cast<size_t*>(&z[i])};
             size_t pi_ng;
             size_t pos;
             // check if this point is new
