@@ -8,25 +8,26 @@ namespace netgen
 {
   struct PointTree
   {
-      BoxTree<3> tree;
+      std::map<int, BoxTree<3>> tree;
+      Box<3> bounding_box;
 
-      PointTree( Box<3> bb ) : tree(bb) {}
+      PointTree( Box<3> bb ) : bounding_box(bb) {}
 
-      void Insert(Point<3> p, PointIndex n)
+      void Insert(Point<3> p, PointIndex n, int index)
       {
-          tree.Insert(p, p, n);
+        if(tree.count(index) == 0)
+          tree.emplace(index, bounding_box);
+        tree.at(index).Insert(p, p, n);
       }
 
-      PointIndex Find(Point<3> p) const
+      PointIndex Find(Point<3> p, int index) const
       {
           ArrayMem<int, 1> points;
-          tree.GetIntersecting(p, p, points);
+          tree.at(index).GetIntersecting(p, p, points);
           if(points.Size()==0)
               throw Exception("cannot find mapped point " + ToString(p));
           return points[0];
       }
-
-      double GetTolerance() { return tree.GetTolerance(); }
   };
 
   DLL_HEADER GeometryRegisterArray geometryregister;
@@ -213,22 +214,24 @@ namespace netgen
       }
   }
 
-  struct Line
-  {
-    Point<3> p0, p1;
-    inline double Length() const { return (p1-p0).Length(); }
-    inline double Dist(const Line& other) const
+  namespace {
+    struct Line
     {
-      Vec<3> n = p1-p0;
-      Vec<3> q = other.p1-other.p0;
-      double nq = n*q;
-      Point<3> p = p0 + 0.5*n;
-      double lambda = (p-other.p0)*n / (nq + 1e-10);
-      if (lambda >= 0 && lambda <= 1)
-        return (p-other.p0-lambda*q).Length();
-      return 1e99;
-    }
-  };
+      Point<3> p0, p1;
+      inline double Length() const { return (p1-p0).Length(); }
+      inline double Dist(const Line& other) const
+      {
+        Vec<3> n = p1-p0;
+        Vec<3> q = other.p1-other.p0;
+        double nq = n*q;
+        Point<3> p = p0 + 0.5*n;
+        double lambda = (p-other.p0)*n / (nq + 1e-10);
+        if (lambda >= 0 && lambda <= 1)
+          return (p-other.p0-lambda*q).Length();
+        return 1e99;
+      }
+    };
+  }
 
   void NetgenGeometry :: Clear()
   {
@@ -581,17 +584,17 @@ namespace netgen
 
     Array<PointIndex> vert2meshpt(vertices.Size());
     vert2meshpt = PointIndex::INVALID;
+
     for(auto & vert : vertices)
       {
         auto pi = mesh.AddPoint(vert->GetPoint(), vert->properties.layer);
-        tree.Insert(mesh[pi], pi);
         vert2meshpt[vert->nr] = pi;
         mesh[pi].Singularity(vert->properties.hpref);
         mesh[pi].SetType(FIXEDPOINT);
-
-        Element0d el(pi, pi);
+        
+        Element0d el(pi, pi-IndexBASE<PointIndex>()+1);
         el.name = vert->properties.GetName();
-        mesh.SetCD3Name(pi, el.name);
+        mesh.SetCD3Name(pi-IndexBASE<PointIndex>()+1, el.name);
         mesh.pointelements.Append (el);
       }
 
@@ -616,8 +619,8 @@ namespace netgen
         endp = vert2meshpt[edge->GetEndVertex().nr];
 
         // ignore collapsed edges
-        if(startp == endp && edge->GetLength() < 1e-10 * bounding_box.Diam())
-            continue;
+        if(edge->IsDegenerated())
+          continue;
 
         // ----------- Add Points to mesh and create segments -----
         auto & pnums = all_pnums[edgenr];
@@ -716,7 +719,8 @@ namespace netgen
         for(auto i : Range(edge_points))
         {
             auto pi = mesh.AddPoint(edge_points[i], edge->properties.layer);
-            tree.Insert(mesh[pi], pi);
+            if(edge->identifications.Size())
+              tree.Insert(mesh[pi], pi, edge->nr);
             pnums[i+1] = pi;
         }
 
@@ -757,7 +761,7 @@ namespace netgen
                   p_other = (*ident.trafo)(mesh[pi]);
                 else
                   static_cast<GeometryEdge*>(ident.to)->ProjectPoint(p_other, nullptr);
-                auto pi_other = tree.Find(p_other);
+                auto pi_other = tree.Find(p_other, ident.to->nr);
                 identifications.Add(pi, pi_other, ident.name, ident.type);
             }
           }
@@ -866,7 +870,7 @@ namespace netgen
                 for(auto pi : s.PNums())
                     if(!is_point_in_tree[pi])
                     {
-                        tree.Insert(mesh[pi], pi);
+                        tree.Insert(mesh[pi], pi, -1);
                         is_point_in_tree[pi] = true;
                     }
 
@@ -936,8 +940,8 @@ namespace netgen
                           edges[mapped_edges[edgenr]]->ProjectPoint(p2, nullptr);
                           edges[mapped_edges[edgenr]]->ProjectPoint(p3, nullptr);
                         }
-                        sel[2] = tree.Find(p2);
-                        sel[3] = tree.Find(p3);
+                        sel[2] = tree.Find(p2, -1);
+                        sel[3] = tree.Find(p3, -1);
 
                         // find mapped segment to set PointGeomInfo correctly
                         Segment s_other;
@@ -1017,7 +1021,7 @@ namespace netgen
                         if(mesh[pi].Type() == SURFACEPOINT && pi_to_face[pi]==-1)
                         {
                             pi_to_face[pi] = face->nr;
-                            tree.Insert(mesh[pi], pi);
+                            tree.Insert(mesh[pi], pi, -1);
                             pi_of_face[face->nr].Append(pi);
                         }
                     }
@@ -1085,7 +1089,7 @@ namespace netgen
                       else
                         edge->ProjectPoint(p, nullptr);
                     }
-              tree.Insert(p, pi);
+              tree.Insert(p, pi, -1);
               is_point_in_tree[pi] = true;
             }
           }
@@ -1097,7 +1101,7 @@ namespace netgen
           {
             auto pi = seg[i];
             if(!pmap[pi].IsValid())
-                pmap[tree.Find(mesh[pi])] = pi;
+              pmap[tree.Find(mesh[pi], -1)] = pi;
 
             // store uv values (might be different values for same point in case of internal edges)
             double u = seg.epgeominfo[i].u;
@@ -1244,8 +1248,7 @@ namespace netgen
 
   void NetgenGeometry :: FinalizeMesh(Mesh& mesh) const
   {
-    if(solids.Size())
-      for (int i = 0; i < mesh.GetNDomains(); i++)
+      for (int i = 0; i < std::min(solids.Size(), (size_t)mesh.GetNDomains()); i++)
         if (auto name = solids[i]->properties.name)
           mesh.SetMaterial (i+1, *name);
 
