@@ -86,6 +86,9 @@ namespace netgen
   protected:
       GeometryVertex *start, *end;
   public:
+    // Neighboring domains in 2d
+    // In 3d unused, EXCEPT for free floating edges in a domain,
+    // then both are pointing to the containing domain
     int domin=-1, domout=-1;
 
     GeometryEdge( GeometryVertex &start_, GeometryVertex &end_ )
@@ -187,7 +190,10 @@ namespace netgen
   };
 
   class DLL_HEADER GeometrySolid : public GeometryShape
-  { };
+  {
+  public:
+    Array<GeometryEdge*> free_edges; // edges with no adjacent face
+  };
 
   class DLL_HEADER NetgenGeometry
   {
@@ -212,11 +218,18 @@ namespace netgen
     size_t GetNVertices() const { return vertices.Size(); }
     size_t GetNEdges() const { return edges.Size(); }
     size_t GetNFaces() const { return faces.Size(); }
+    size_t GetNSolids() const { return solids.Size(); }
 
+    const GeometrySolid & GetSolid(int i) const { return *solids[i]; }
     const GeometryFace & GetFace(int i) const { return *faces[i]; }
     const GeometryEdge & GetEdge(int i) const { return *edges[i]; }
     const GeometryVertex & GetVertex(int i) const { return *vertices[i]; }
 
+    auto Solids() const { return FlatArray{solids}; }
+    auto Faces() const { return FlatArray{faces}; }
+    auto Edges() const { return FlatArray{edges}; }
+    auto Vertices() const { return FlatArray{vertices}; }
+
     virtual Array<const GeometryVertex*> GetFaceVertices(const GeometryFace& face) const { return Array<const GeometryVertex*>{}; }
 
     void Clear();
@@ -258,7 +271,7 @@ namespace netgen
 
   virtual void ProjectPointEdge (int surfind, int surfind2, Point<3> & p, EdgePointGeomInfo* gi = nullptr) const
   {
-    if(gi && gi->edgenr < edges.Size())
+    if(gi && gi->edgenr < edges.Size() && gi->edgenr >= 0)
       edges[gi->edgenr]->ProjectPoint(p, gi);
   }
 
