@@ -1,5 +1,5 @@
 #include <mystdlib.h>
-#include "meshing.hpp"
+#include "boundarylayer.hpp"
 #include "meshing2.hpp"
 #include "../geom2d/csg2d.hpp"
 
@@ -283,7 +283,7 @@ namespace netgen
     }
 
     {
-        FaceDescriptor new_fd(0, 0, 0, -1);
+      FaceDescriptor new_fd(0, 0, 0, -1);
         new_fd.SetBCProperty(new_domain);
         // int new_fd_index =
         mesh.AddFaceDescriptor(new_fd);
@@ -329,7 +329,7 @@ namespace netgen
         auto current_si = si;
 
         auto first = current_seg[0];
-        auto current = -1;
+        PointIndex current(PointIndex::INVALID);
         auto next =  current_seg[1];
 
         if(points_done.Test(first))
@@ -354,7 +354,7 @@ namespace netgen
                  current_si = sj;
                  current_seg = mesh[sj];
 
-                 next = current_seg[0] + current_seg[1] - current;
+                 next = current_seg[0]-current + current_seg[1];
                  break;
               }
            }
@@ -493,7 +493,7 @@ namespace netgen
            if(growthvectors[pi].Length2() == 0.0)
               continue;
 
-           PointIndex pi1 = seg0[0] + seg0[1] - pi;
+           PointIndex pi1 = seg0[0] - pi + seg0[1];
            auto p1 = mesh[pi1];
            auto p = mesh[pi];
 
@@ -584,13 +584,15 @@ namespace netgen
         auto p2 = [](Point<3> p) { return Point<2>{p[0], p[1]}; };
 
         auto seg = line_segments[segi];
-        double alpha,beta;
-        intersect( p2(mesh[seg[0]]), p2(mesh[seg[0]]+total_thickness*growthvectors[seg[0]]), p2(mesh[seg[1]]), p2(mesh[seg[1]]+total_thickness*growthvectors[seg[1]]), alpha, beta );
-
-        if(beta>0 && alpha>0 && alpha<1.1)
-           growth[seg[0]] = min(growth[seg[0]], 0.8*alpha);
-        if(alpha>0 && beta>0 && beta<1.1)
-           growth[seg[1]] = min(growth[seg[1]], 0.8*beta);
+        double alpha=0.0;
+        double beta=0.0;
+        if (intersect(p2(mesh[seg[0]]), p2(mesh[seg[0]] + total_thickness * growthvectors[seg[0]]), p2(mesh[seg[1]]), p2(mesh[seg[1]] + total_thickness * growthvectors[seg[1]]), alpha, beta))
+          {
+            if (beta > 0 && alpha > 0 && alpha < 1.1)
+              growth[seg[0]] = min(growth[seg[0]], 0.8 * alpha);
+            if (alpha > 0 && beta > 0 && beta < 1.1)
+              growth[seg[1]] = min(growth[seg[1]], 0.8 * beta);
+          }
 
         for (auto segj : Range(mesh.LineSegments()))
            if(segi!=segj)
