@@ -603,19 +603,21 @@ namespace netgen
     return hps;
   }
 
-  bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HASHTABLE<int> & edgepoiclt_dom, 
+  template <typename HT_EDGEPOINT_DOM>
+  bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, HT_EDGEPOINT_DOM & edgepoiclt_dom, 
 		       NgBitArray & cornerpoint, NgBitArray & edgepoint, INDEX_3_HASHTABLE<int> & faces, INDEX_2_HASHTABLE<int> & face_edges, 
-			INDEX_2_HASHTABLE<int> & surf_edges, NgArray<int, PointIndex::BASE> & facepoint, int & levels, int & act_ref); 
+			INDEX_2_HASHTABLE<int> & surf_edges, Array<int, PointIndex> & facepoint, int & levels, int & act_ref); 
 
   bool ClassifyHPElements (Mesh & mesh, NgArray<HPRefElement> & elements, SplittingType split, int & act_ref, int & levels);
   
   
   void  InitHPElements(Mesh & mesh, NgArray<HPRefElement> & elements) 
   { 
-    for(ElementIndex i = 0; i < mesh.GetNE(); i++) 
+    // for(ElementIndex i = 0; i < mesh.GetNE(); i++)
+    for(ElementIndex i : mesh.VolumeElements().Range())
       {
 	HPRefElement hpel(mesh[i]); 
-	hpel.coarse_elnr = i; 
+	hpel.coarse_elnr = int(i); 
 	
 	switch (mesh[i].GetType()) 
 	  { 
@@ -699,6 +701,8 @@ namespace netgen
 	  {
 	    INDEX_2 i2(el.pnums[hprs->splitedges[j][0]-1],
 		       el.pnums[hprs->splitedges[j][1]-1]);
+            if (fac1 == 0.5) i2.Sort();
+            
 	    if (!newpts.Used (i2))
 	      {
 		Point<3> np; 
@@ -787,7 +791,8 @@ namespace netgen
 	  {
 	    INDEX_2 i2(el.pnums[hprs->splitedges[j][0]-1],
 		       el.pnums[hprs->splitedges[j][1]-1]);
-
+            if (fac1 == 0.5) i2.Sort();
+            
 	    int npi = newpts.Get(i2);
 	    newpnums[hprs->splitedges[j][2]-1] = npi;
 
@@ -1506,7 +1511,8 @@ namespace netgen
 
     if(act_ref>=1)
       { 
-	for(ElementIndex i=0;i<mesh.GetNE(); i++) 
+	// for(ElementIndex i=0;i<mesh.GetNE(); i++)
+        for (ElementIndex i : mesh.VolumeElements().Range())
 	  { 
 	    // Element el = mesh[i] ;
 	    HPRefElement & hpel = hpelements[mesh[i].GetHpElnr()];
@@ -1629,9 +1635,10 @@ namespace netgen
       }
   }
 
-bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HASHTABLE<int> & edgepoint_dom, 
-		       NgBitArray & cornerpoint, NgBitArray & edgepoint, INDEX_3_HASHTABLE<int> & faces, INDEX_2_HASHTABLE<int> & face_edges, 
-			INDEX_2_HASHTABLE<int> & surf_edges, NgArray<int, PointIndex::BASE> & facepoint, int & levels, int & act_ref)
+  template <typename HT_EDGEPOINT_DOM>
+  bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, HT_EDGEPOINT_DOM & edgepoint_dom, 
+		       TBitArray<PointIndex> & cornerpoint, TBitArray<PointIndex> & edgepoint, INDEX_3_HASHTABLE<int> & faces, INDEX_2_HASHTABLE<int> & face_edges, 
+			INDEX_2_HASHTABLE<int> & surf_edges, Array<int, PointIndex> & facepoint, int & levels, int & act_ref)
 {
   bool sing = 0; 
   if (mesh.GetDimension() == 3)
@@ -1667,7 +1674,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	  {
 	    if (mesh.Point(i).Singularity() * levels >= act_ref)
 	      {
-		cornerpoint.Set(i);
+		cornerpoint.SetBit(i);
 		sing = 1; 
 	      } 
 	  }
@@ -1692,16 +1699,17 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	      edges.Set (i2s, 1);
 
 
-	      edgepoint.Set (i2.I1());
-	      edgepoint.Set (i2.I2());
+	      edgepoint.SetBit (i2.I1());
+	      edgepoint.SetBit (i2.I2());
 	      sing = 1; 
 	    }
 
 	// if 2 adjacent edges of an element are singular, the 
 	// common point must be a singular point
-	for (int i = 1; i <= mesh.GetNE(); i++)
+	// for (int i = 1; i <= mesh.GetNE(); i++)
+        for (auto ei : mesh.VolumeElements().Range())
 	  {
-	    const Element & el = mesh.VolumeElement(i);
+	    const Element & el = mesh[ei]; 
 	    const ELEMENT_EDGE * eledges = MeshTopology::GetEdges1 (el.GetType());
 	    int nedges = MeshTopology::GetNEdges (el.GetType());
 	    for (int j = 0; j < nedges; j++)
@@ -1714,10 +1722,10 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 		    ek.Sort();
 		    if (edges.Used(ej) && edges.Used(ek))
 		      {
-			if (ej.I1() == ek.I1()) cornerpoint.Set (ek.I1());
-			if (ej.I1() == ek.I2()) cornerpoint.Set (ek.I2());
-			if (ej.I2() == ek.I1()) cornerpoint.Set (ek.I1());
-			if (ej.I2() == ek.I2()) cornerpoint.Set (ek.I2());
+			if (ej.I1() == ek.I1()) cornerpoint.SetBit (ek.I1());
+			if (ej.I1() == ek.I2()) cornerpoint.SetBit (ek.I2());
+			if (ej.I2() == ek.I1()) cornerpoint.SetBit (ek.I1());
+			if (ej.I2() == ek.I2()) cornerpoint.SetBit (ek.I2());
 		      }
 		  }
 	  }
@@ -1797,33 +1805,33 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    if (seg.singedge_left * levels >= act_ref)
 	      {
 		INDEX_2 i2 = INDEX_2::Sort(mesh.LineSegment(i)[0], 
-			    mesh.LineSegment(i)[1]);
+                                           mesh.LineSegment(i)[1]);
 		edges.Set(i2,1); 
-		edgepoint.Set(i2.I1());
-		edgepoint.Set(i2.I2());
+		edgepoint.SetBit(i2.I1());
+		edgepoint.SetBit(i2.I2());
 		*testout << " singleft " << endl;  
 		*testout << " mesh.LineSegment(i).domout " << mesh.LineSegment(i).domout << endl;      
 		*testout << " mesh.LineSegment(i).domin " << mesh.LineSegment(i).domin << endl;      
-		edgepoint_dom.Set (INDEX_2(mesh.LineSegment(i).domin, i2.I1()), 1);
-		edgepoint_dom.Set (INDEX_2(mesh.LineSegment(i).domin, i2.I2()), 1);
+		edgepoint_dom.Set ( { mesh.LineSegment(i).domin, i2.I1() }, 1);
+		edgepoint_dom.Set ( { mesh.LineSegment(i).domin, i2.I2() }, 1);
 		sing = 1; 
 		
 	      }
 	    
 	    if (seg.singedge_right * levels >= act_ref)
 	      {
-		INDEX_2 i2 = INDEX_2::Sort(mesh.LineSegment(i)[1], 
-			    mesh.LineSegment(i)[0]);  
+		PointIndices<2> i2 = INDEX_2::Sort(mesh.LineSegment(i)[1], 
+                                                   mesh.LineSegment(i)[0]);  
 		edges.Set (i2, 1);
-		edgepoint.Set(i2.I1());
-		edgepoint.Set(i2.I2());
+		edgepoint.SetBit(i2.I1());
+		edgepoint.SetBit(i2.I2());
 		
 		*testout << " singright " << endl;  
 		*testout << " mesh.LineSegment(i).domout " << mesh.LineSegment(i).domout << endl;      
 		*testout << " mesh.LineSegment(i).domin " << mesh.LineSegment(i).domin << endl;      
 		
-		edgepoint_dom.Set (INDEX_2(mesh.LineSegment(i).domout, i2.I1()), 1);
-		edgepoint_dom.Set (INDEX_2(mesh.LineSegment(i).domout, i2.I2()), 1);
+		edgepoint_dom.Set ( { mesh.LineSegment(i).domout, i2.I1() }, 1);
+		edgepoint_dom.Set ( { mesh.LineSegment(i).domout, i2.I2() }, 1);
 		sing = 1;
 	      }
 	    
@@ -1854,14 +1862,14 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    if (surfonpoint.Get(i).I1())
 	      {
 		// cornerpoint.Set(i);    // disabled by JS, Aug 2009
-		edgepoint.Set(i);
+		edgepoint.SetBit(i);
 	      }
 	
 	    // mark points for refinement that are explicitly specified in input file
 	    if (mesh.Point(i).Singularity()*levels >= act_ref)
 	      {
-		cornerpoint.Set(i);
-		edgepoint.Set(i);
+		cornerpoint.SetBit(i);
+		edgepoint.SetBit(i);
 		sing =  1; 
 	      }
 	  }
@@ -1886,11 +1894,13 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
   bool ClassifyHPElements (Mesh & mesh, NgArray<HPRefElement> & elements, SplittingType split, int & act_ref, int & levels)
   {
     INDEX_2_HASHTABLE<int> edges(mesh.GetNSeg()+1);
-    NgBitArray edgepoint(mesh.GetNP());
-    INDEX_2_HASHTABLE<int> edgepoint_dom(mesh.GetNSeg()+1);
+    TBitArray<PointIndex> edgepoint(mesh.GetNP());
+    // INDEX_2_HASHTABLE<int> edgepoint_dom(mesh.GetNSeg()+1);
 
+    HT_EDGEPOINT_DOM edgepoint_dom;
+    
     edgepoint.Clear();
-    NgBitArray cornerpoint(mesh.GetNP());
+    TBitArray<PointIndex> cornerpoint(mesh.GetNP());
     cornerpoint.Clear();
 
     // value = nr > 0 ... refine elements in domain nr
@@ -1898,7 +1908,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
     INDEX_3_HASHTABLE<int> faces(mesh.GetNSE()+1);
     INDEX_2_HASHTABLE<int> face_edges(mesh.GetNSE()+1);
     INDEX_2_HASHTABLE<int> surf_edges(mesh.GetNSE()+1);
-    NgArray<int, PointIndex::BASE> facepoint(mesh.GetNP());
+    Array<int, PointIndex> facepoint(mesh.GetNP());
 
     bool sing = CheckSingularities(mesh, edges, edgepoint_dom, 
 			      cornerpoint, edgepoint, faces, face_edges, 
@@ -1914,7 +1924,7 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
     NgArray<int> misses(10000);
     misses = 0;
 
-    (*testout) << "edgepoint_dom = " << endl << edgepoint_dom << endl;
+    // (*testout) << "edgepoint_dom = " << endl << edgepoint_dom << endl;
 
     
     for( int i = 0; i<elements.Size(); i++) 
@@ -1990,8 +2000,11 @@ bool CheckSingularities(Mesh & mesh, INDEX_2_HASHTABLE<int> & edges, INDEX_2_HAS
 	    }
 	  case HP_SEGM: 
 	    {
-	      hpel.type = ClassifySegm(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, 
-				       faces, face_edges, surf_edges, facepoint);    
+              if (split == SPLIT_HP)
+                hpel.type = ClassifySegm(hpel, edges, edgepoint_dom, cornerpoint, edgepoint, 
+                                         faces, face_edges, surf_edges, facepoint);
+              else if (split == SPLIT_POWELL)
+                hpel.type = HP_SEGM_SINGCORNERL;
 	      dd = 1; 
 	      break; 
 	    }
