@@ -4,10 +4,6 @@
 namespace netgen
 {
 
-extern double CalcTotalBad (const Mesh::T_POINTS & points, 
-			    const Array<Element, ElementIndex> & elements,
-			    const MeshingParameters & mp);
-
 
 ///
 class MeshOptimize3d
@@ -39,18 +35,18 @@ public:
   void CombineImprove ();
 
   void SplitImprove ();
-  double SplitImproveEdge (Table<ElementIndex,PointIndex> & elementsonnode, NgArray<INDEX_3> &locfaces, double badmax, PointIndex pi1, PointIndex pi2, PointIndex ptmp, bool check_only=false);
+  double SplitImproveEdge (Table<ElementIndex,PointIndex> & elementsonnode, NgArray<PointIndices<3>> &locfaces, double badmax, PointIndex pi1, PointIndex pi2, PointIndex ptmp, bool check_only=false);
 
   void SplitImprove2 ();
   double SplitImprove2Element (ElementIndex ei, const Table<ElementIndex, PointIndex> & elements_of_point, bool check_only);
   
 
-  double SwapImproveEdge (const NgBitArray * working_elements, Table<ElementIndex,PointIndex> & elementsonnode, INDEX_3_HASHTABLE<int> & faces, PointIndex pi1, PointIndex pi2, bool check_only=false);
-  void SwapImprove (const NgBitArray * working_elements = NULL);
-  void SwapImproveSurface (const NgBitArray * working_elements = NULL,
-			   const NgArray< NgArray<int,PointIndex::BASE>* > * idmaps = NULL);
+  double SwapImproveEdge (const TBitArray<ElementIndex> * working_elements, Table<ElementIndex,PointIndex> & elementsonnode, INDEX_3_HASHTABLE<int> & faces, PointIndex pi1, PointIndex pi2, bool check_only=false);
+  void SwapImprove (const TBitArray<ElementIndex> * working_elements = NULL);
+  void SwapImproveSurface (const TBitArray<ElementIndex> * working_elements = NULL,
+			   const NgArray< idmap_type* > * idmaps = NULL);
   void SwapImprove2 ();
-  double SwapImprove2 (ElementIndex eli1, int face, Table<ElementIndex, PointIndex> & elementsonnode, TABLE<SurfaceElementIndex, PointIndex::BASE> & belementsonnode, bool check_only=false );
+  double SwapImprove2 (ElementIndex eli1, int face, Table<ElementIndex, PointIndex> & elementsonnode, DynamicTable<SurfaceElementIndex, PointIndex> & belementsonnode, bool check_only=false );
 
   void ImproveMesh() { mesh.ImproveMesh(mp, goal); }
 
@@ -64,13 +60,6 @@ public:
   }
 
 
-  double CalcTotalBad (const Mesh::T_POINTS & points, 
-		       const Array<Element, ElementIndex> & elements)
-  {
-    return netgen::CalcTotalBad (points, elements, mp);
-  }
-
-
   double GetLegalPenalty()
   {
     return goal == OPT_LEGAL ? 1e15 : 1e6;
@@ -119,12 +108,12 @@ public:
 class PointFunction1 : public MinFunction
 {
   Mesh::T_POINTS & points;
-  const NgArray<INDEX_3> & faces;
+  const NgArray<PointIndices<3>> & faces;
   const MeshingParameters & mp;
   double h;
 public:
   PointFunction1 (Mesh::T_POINTS & apoints, 
-		  const NgArray<INDEX_3> & afaces,
+		  const NgArray<PointIndices<3>> & afaces,
 		  const MeshingParameters & amp,
 		  double ah);
   
