@@ -144,7 +144,40 @@ namespace netgen
 #endif
 
 
+#ifdef PARALLEL
+  NG_MPI_Datatype Element0d :: MyGetMPIType()
+  {
+    static NG_MPI_Datatype type = NG_MPI_DATATYPE_NULL;
+    static NG_MPI_Datatype htype = NG_MPI_DATATYPE_NULL;
+    if (type == NG_MPI_DATATYPE_NULL)
+      {
+	Element0d hel;
+	int blocklen[] = { 1, 1 };
+	NG_MPI_Aint displ[] =
+          { (char*)&hel.pnum - (char*)&hel,
+            (char*)&hel.index - (char*)&hel,
+          };
+	NG_MPI_Datatype types[] = {
+          GetMPIType(hel.pnum), GetMPIType(hel.index)
+        };
+	NG_MPI_Type_create_struct (2, blocklen, displ, types, &htype);
+	NG_MPI_Type_commit ( &htype );
+	NG_MPI_Aint lb, ext;
+	NG_MPI_Type_get_extent (htype, &lb, &ext);
+	// *testout << "lb = " << lb << endl;
+	// *testout << "ext = " << ext << endl;
+	ext = sizeof (Element0d);
+	NG_MPI_Type_create_resized (htype, lb, ext, &type);
+	NG_MPI_Type_commit ( &type );
+      }
+    return type;
+  }
+#endif
 
+ void Element0d :: DoArchive (Archive & ar)
+ {
+   ar & pnum & index;
+ }
 
   Segment :: Segment() 
     : is_curved(false)
@@ -204,7 +237,7 @@ namespace netgen
   {
     for (int i = 0; i < ELEMENT2D_MAXPOINTS; i++)
       {
-	pnum[i] = 0;
+	pnum[i].Invalidate();
 	geominfo[i].trignum = 0;
       }
     np = 3;
@@ -266,7 +299,7 @@ namespace netgen
 
 
 
-  Element2d :: Element2d (int pi1, int pi2, int pi3)
+  Element2d :: Element2d (PointIndex pi1, PointIndex pi2, PointIndex pi3)
   {
     pnum[0] = pi1;
     pnum[1] = pi2;
@@ -289,7 +322,7 @@ namespace netgen
     is_curved = false;
   }
 
-  Element2d :: Element2d (int pi1, int pi2, int pi3, int pi4)
+  Element2d :: Element2d (PointIndex pi1, PointIndex pi2, PointIndex pi3, PointIndex pi4)
   {
     pnum[0] = pi1;
     pnum[1] = pi2;
@@ -298,8 +331,8 @@ namespace netgen
     np = 4;
     typ = QUAD;
 
-    pnum[4] = 0;
-    pnum[5] = 0;
+    pnum[4].Invalidate();
+    pnum[5].Invalidate();
   
     for (int i = 0; i < ELEMENT2D_MAXPOINTS; i++)
       geominfo[i].trignum = 0;
@@ -1005,7 +1038,7 @@ namespace netgen
   {
     s << "np = " << el.GetNP();
     for (int j = 0; j < el.GetNP(); j++)
-      s << " " << int(el[j]);
+      s << " " << el[j];
     return s;
   }
 
@@ -2661,6 +2694,7 @@ namespace netgen
   {
     identifiedpoints.DeleteData();
     identifiedpoints_nr.DeleteData();
+    idpoints_table.SetSize(0);
 
     /*
     delete identifiedpoints;
@@ -2675,8 +2709,8 @@ namespace netgen
     {
       ar & maxidentnr;
       ar & identifiedpoints & identifiedpoints_nr;
-
       ar & idpoints_table;
+      
       if (ar.Output())
         {
           size_t s = type.Size();
@@ -2692,6 +2726,8 @@ namespace netgen
           for (auto & t : type)
             ar & (unsigned char&)(t);
         }
+      if (ar.GetVersion("netgen") > "v6.2.2404-66")
+        ar & names;
     }    
 
   
@@ -2699,11 +2735,11 @@ namespace netgen
   void Identifications :: Add (PointIndex pi1, PointIndex pi2, int identnr)
   {
     //  (*testout) << "Identification::Add, pi1 = " << pi1 << ", pi2 = " << pi2 << ", identnr = " << identnr << endl;
-    INDEX_2 pair (pi1, pi2);
+    PointIndices<2> pair (pi1, pi2);
     identifiedpoints.Set (pair, identnr);
 
-    INDEX_3 tripl (pi1, pi2, identnr);
-    identifiedpoints_nr.Set (tripl, 1);
+    // INDEX_3 tripl (pi1, pi2, identnr);
+    identifiedpoints_nr.Set ( { { pi1, pi2 }, identnr }, 1);
 
     if (identnr > maxidentnr) maxidentnr = identnr;
     names.SetSize(maxidentnr);
@@ -2726,8 +2762,9 @@ namespace netgen
 
   bool Identifications :: Get (PointIndex pi1, PointIndex pi2, int nr) const
   {
-    INDEX_3 tripl(pi1, pi2, nr);
-    if (identifiedpoints_nr.Used (tripl))
+    // INDEX_3 tripl(pi1, pi2, nr);
+    // if (identifiedpoints_nr.Used (tripl))
+    if (identifiedpoints_nr.Used ( { { pi1, pi1 }, nr } ) )
       return 1;
     else
       return 0;
@@ -2749,7 +2786,7 @@ namespace netgen
   }
 
 
-  void Identifications :: GetMap (int identnr, NgArray<int,PointIndex::BASE> & identmap, bool symmetric) const
+  void Identifications :: GetMap (int identnr, idmap_type & identmap, bool symmetric) const
   {
     identmap.SetSize (mesh.GetNP());
     identmap = 0;
@@ -2767,18 +2804,30 @@ namespace netgen
       {
         cout << "getmap, identnr = " << identnr << endl;
 
+        /*
         for (int i = 1; i <= identifiedpoints_nr.GetNBags(); i++)
           for (int j = 1; j <= identifiedpoints_nr.GetBagSize(i); j++)
+        */
+        for (auto [hash, val] : identifiedpoints_nr)
             {
+              /*
               INDEX_3 i3;
               int dummy;
               identifiedpoints_nr.GetData (i, j, i3, dummy);
-	    
-              if (i3.I3() == identnr || !identnr)
+              */
+
+              auto [hash_pts, hash_nr] = hash;
+              
+              if (hash_nr == identnr || !identnr)
                 {
+                  /*
                   identmap.Elem(i3.I1()) = i3.I2();
                   if(symmetric)
                     identmap.Elem(i3.I2()) = i3.I1();
+                  */
+                  identmap[hash_pts.I1()] = hash_pts.I2();
+                  if(symmetric)
+                    identmap[hash_pts.I2()] = hash_pts.I1();
                 }
             }  
       }
@@ -2786,12 +2835,26 @@ namespace netgen
   }
 
 
+  Array<INDEX_3> Identifications :: GetPairs () const
+  {
+    Array<INDEX_3> pairs;
+    for(auto [hash, dummy] : identifiedpoints_nr)
+      // pairs.Append(i3);
+      {
+        auto [pts,nr] = hash;
+        pairs.Append ( { pts[0], pts[1], nr } );
+      }
+    return pairs;
+  }
+
   void Identifications :: GetPairs (int identnr, 
                                     NgArray<INDEX_2> & identpairs) const
   {
     identpairs.SetSize(0);
   
     if (identnr == 0)
+      {
+        /*
       for (int i = 1; i <= identifiedpoints.GetNBags(); i++)
         for (int j = 1; j <= identifiedpoints.GetBagSize(i); j++)
           {
@@ -2799,8 +2862,14 @@ namespace netgen
             int nr;
             identifiedpoints.GetData (i, j, i2, nr);
             identpairs.Append (i2);
-          }  
+          }
+        */
+        for (auto [hash,val] : identifiedpoints)
+          identpairs.Append (hash);
+      }
     else
+      {
+        /*
       for (int i = 1; i <= identifiedpoints_nr.GetNBags(); i++)
         for (int j = 1; j <= identifiedpoints_nr.GetBagSize(i); j++)
           {
@@ -2810,12 +2879,21 @@ namespace netgen
 	  
             if (i3.I3() == identnr)
               identpairs.Append (INDEX_2(i3.I1(), i3.I2()));
-          }  
+          }
+        */
+        for (auto [hash,val] : identifiedpoints_nr)
+          {
+            auto [hash_pts, hash_nr] = hash;
+            if (hash_nr == identnr)
+              identpairs.Append (hash_pts);
+          }
+      }
   }
 
 
   void Identifications :: SetMaxPointNr (int maxpnum)
   {
+    /*
     for (int i = 1; i <= identifiedpoints.GetNBags(); i++)
       for (int j = 1; j <= identifiedpoints.GetBagSize(i); j++)
         {
@@ -2829,17 +2907,105 @@ namespace netgen
               identifiedpoints.SetData (i, j, i2, -1);	    
             }
         }
+    */
+
+    // can we get data by reference ? 
+    for (auto [hash,data] : identifiedpoints)
+      {
+        if (hash.I1() > IndexBASE<PointIndex>()+maxpnum-1 ||
+            hash.I2() > IndexBASE<PointIndex>()+maxpnum-1)
+          {
+            identifiedpoints[hash] = -1;
+          }
+            
+      }
   }
 
+  // Map points in the identifications to new point numbers
+  // deletes identifications with invalid (mapped) points
+  void Identifications :: MapPoints(FlatArray<PointIndex, PointIndex> op2np)
+  {
+    auto pairs = GetPairs();
+    Delete();
+
+    for(auto pair : pairs)
+      {
+        auto p1 = op2np[pair.I1()];
+        auto p2 = op2np[pair.I2()];
+        if(p1.IsValid() && p2.IsValid())
+          Add(p1, p2, pair.I3());
+      }
+  }
 
   void Identifications :: Print (ostream & ost) const
   {
     ost << "Identifications:" << endl;
     ost << "pairs: " << endl << identifiedpoints << endl;
-    ost << "pairs and nr: " << endl << identifiedpoints_nr << endl;
+    // ost << "pairs and nr: " << endl << identifiedpoints_nr << endl;
+#pragma message( "Can't ostream a tuple " __FILE__ )    
     ost << "table: " << endl << idpoints_table << endl;
   }
 
+  ostream & operator<< (ostream & ost, const BoundaryLayerParameters & mp)
+  {
+    auto print_vec = [&ost](auto & v) {
+        ost << "[";
+        for (const auto & val : v)
+          ost << val << " ";
+        ost << "]";
+    };
+    ost << "BoundaryLayerParameters" << endl;
+    ost << "  boundary: ";
+    switch(mp.boundary.index())
+      {
+        case 0: ost << std::get<0>(mp.boundary); break;
+        case 1: ost << std::get<1>(mp.boundary); break;
+        case 2: print_vec(std::get<2>(mp.boundary)); break;
+      }
+    ost << "\n  thickness: ";
+    switch(mp.thickness.index())
+      {
+        case 0: ost << std::get<0>(mp.thickness); break;
+        case 1: print_vec(std::get<1>(mp.thickness)); break;
+      }
+    ost <<"\n  new_material: ";
+    if(!mp.new_material) ost << "nullopt";
+    else {
+      switch(mp.new_material->index())
+        {
+          case 0: ost << std::get<0>(*mp.new_material); break;
+          case 1:
+          for (const auto & [key, value] : std::get<1>(*mp.new_material))
+            ost << key << " -> " << value << ", ";
+          break;
+        }
+    }
+    ost << "\n  domain: ";
+    switch(mp.domain.index())
+      {
+        case 0: ost << std::get<0>(mp.domain); break;
+        case 1: ost << std::get<1>(mp.domain); break;
+        case 2: print_vec(std::get<2>(mp.domain)); break;
+      }
+    ost << "\n  outside: " << mp.outside;
+    ost << "\n  project_boundaries: ";
+    if(mp.project_boundaries) {
+      auto & proj_bnd = *mp.project_boundaries;
+      switch(proj_bnd.index())
+        {
+          case 0: ost << std::get<0>(proj_bnd); break;
+          case 1: print_vec(std::get<1>(proj_bnd)); break;
+        }
+    }
+    else
+      ost << "nullopt";
+    ost << "\n  grow_edges: " << mp.grow_edges;
+    ost << "\n  limit_growth_vectors: " << mp.limit_growth_vectors;
+    ost << "\n  sides_keep_surfaceindex: " << (mp.sides_keep_surfaceindex ? ToString(*mp.sides_keep_surfaceindex) : "nullopt");
+    ost << "\n  disable_curving: " << mp.disable_curving;
+    ost << endl;
+    return ost;
+  }
 
   MeshingParameters :: MeshingParameters ()
   {
