@@ -1,6 +1,3 @@
-// #ifdef PARALLEL
-
-
 #include <meshing.hpp>
 #include "paralleltop.hpp"
 
@@ -96,7 +93,7 @@ namespace netgen
     comm.AllGather (num_master_points, first_master_point);
     auto max_oldv = comm.AllReduce (Max (glob_vert.Range(0, oldnv)), NG_MPI_MAX);
     if (comm.AllReduce (oldnv, NG_MPI_SUM) == 0)
-      max_oldv = PointIndex::BASE-1;
+      max_oldv = long(PointIndex::BASE)-1;
     
     size_t num_glob_points = max_oldv+1;
     for (int i = 0; i < comm.Size(); i++)
@@ -138,16 +135,17 @@ namespace netgen
           for (auto p : dps)
             send_data[p][nsend[p]++] = L2G(pi);
 
-    Array<NG_MPI_Request> requests;
+    NgMPI_Requests requests;
     for (int i = 0; i < comm.Size(); i++)
       {
         if (nsend[i])
-          requests.Append (comm.ISend (send_data[i], i, 200));
+          requests += comm.ISend (send_data[i], i, 200);
         if (nrecv[i])
-          requests.Append (comm.IRecv (recv_data[i], i, 200));
+          requests += comm.IRecv (recv_data[i], i, 200);
       }
     
-    MyMPI_WaitAll (requests);
+    // MyMPI_WaitAll (requests);
+    requests.WaitAll();
     
     Array<int> cnt(comm.Size());
     cnt = 0;
@@ -186,7 +184,7 @@ namespace netgen
 
         if (mesh.mlbetweennodes.Size() == mesh.Points().Size())
           {
-            NgArray<PointIndices<2>,PointIndex::BASE> hml { mesh.mlbetweennodes };
+            Array<PointIndices<2>,PointIndex> hml { mesh.mlbetweennodes };
             for (PointIndex pi : Range(mesh.Points()))
               mesh.mlbetweennodes[inv_index[pi]] = hml[pi];
           }
@@ -292,7 +290,7 @@ namespace netgen
 
 
 
-
+  /*
 
   void ParallelMeshTopology :: UpdateCoarseGridGlobal ()
   {
@@ -387,7 +385,8 @@ namespace netgen
     
     is_updated = true;
   }
-
+  */
+  
   
   void ParallelMeshTopology :: IdentifyVerticesAfterRefinement()
   {
@@ -437,7 +436,7 @@ namespace netgen
 	      for (int dist : GetDistantProcs(pi))
 		dest2vert.Add (dist, pi);
             
-	    for (PointIndex pi = PointIndex::BASE; pi < newnv+PointIndex::BASE; pi++)
+	    for (PointIndex pi = IndexBASE<PointIndex>(); pi < newnv+IndexBASE<PointIndex>(); pi++)
               if (auto [v1,v2] = mesh.mlbetweennodes[pi]; v1.IsValid())              
                 {
                   auto procs1 = GetDistantProcs(v1);
@@ -500,7 +499,6 @@ namespace netgen
 		}
 
 	    DynamicTable<int> recv_verts(ntasks);
-	    // MyMPI_ExchangeTable (send_verts, recv_verts, NG_MPI_TAG_MESH+9, comm);
             comm.ExchangeTable (send_verts, recv_verts, NG_MPI_TAG_MESH+9);
 
 	    for (int dest = 0; dest < ntasks; dest++)
@@ -693,12 +691,8 @@ namespace netgen
           }
       }
 
-    // cout << "UpdateCoarseGrid - edges mpi-exchange" << endl;
-    // TABLE<int> recv_edges(ntasks);
     DynamicTable<int> recv_edges(ntasks);
-    // MyMPI_ExchangeTable (send_edges, recv_edges, NG_MPI_TAG_MESH+9, comm);
     comm.ExchangeTable (send_edges, recv_edges, NG_MPI_TAG_MESH+9);
-    // cout << "UpdateCoarseGrid - edges mpi-exchange done" << endl;
 
     for (int dest = 0; dest < ntasks; dest++)
       {
@@ -803,12 +797,8 @@ namespace netgen
 		}
 	    }
 	
-	// cout << "UpdateCoarseGrid - faces mpi-exchange" << endl;
-	// TABLE<int> recv_faces(ntasks);
 	DynamicTable<int> recv_faces(ntasks);
-	// MyMPI_ExchangeTable (send_faces, recv_faces, NG_MPI_TAG_MESH+9, comm);
         comm.ExchangeTable (send_faces, recv_faces, NG_MPI_TAG_MESH+9);
-	// cout << "UpdateCoarseGrid - faces mpi-exchange done" << endl;
 	
 	for (int dest = 0; dest < ntasks; dest++)
 	  {
@@ -845,4 +835,3 @@ namespace netgen
 }
 
 
-// #endif
