@@ -5,6 +5,7 @@
 #include <BRepTools.hxx>
 
 #include "occ_utils.hpp"
+#include "occgeom.hpp"
 
 namespace netgen
 {
@@ -61,14 +62,14 @@ namespace netgen
     Standard_Integer BuildTriangulation( const TopoDS_Shape & shape )
     {
        BRepTools::Clean (shape);
-       double deflection = 0.01;
+       // double deflection = 0.01;
 
        // https://dev.opencascade.org/doc/overview/html/occt_user_guides__mesh.html
        // from Standard_Boolean meshing_imeshtools_parameters()
        IMeshTools_Parameters aMeshParams;
        aMeshParams.Deflection               = 0.01;
        aMeshParams.Angle                    = 0.5;
-       aMeshParams.Relative                 = Standard_False;
+       aMeshParams.Relative                 = Standard_True;
        aMeshParams.InParallel               = Standard_True;
        aMeshParams.MinSize                  = Precision::Confusion();
        aMeshParams.InternalVerticesMode     = Standard_True;
