@@ -232,6 +232,7 @@ namespace netgen
     using NetgenGeometry::GetVertex;
     using NetgenGeometry::GetEdge;
     using NetgenGeometry::GetFace;
+    using NetgenGeometry::GetSolid;
 
     GeometryShape & GetShape(const TopoDS_Shape & shape)
     {
@@ -252,10 +253,16 @@ namespace netgen
         return const_cast<GeometryFace&>(as_const(*this).GetFace(shape));
     }
 
+    GeometrySolid & GetSolid(const TopoDS_Shape & shape)
+    {
+        return const_cast<GeometrySolid&>(as_const(*this).GetSolid(shape));
+    }
+
     const GeometryShape & GetShape(const TopoDS_Shape & shape) const;
     const GeometryVertex & GetVertex(const TopoDS_Shape & shape) const;
     const GeometryEdge & GetEdge(const TopoDS_Shape & shape) const;
     const GeometryFace & GetFace(const TopoDS_Shape & shape) const;
+    const GeometrySolid & GetSolid(const TopoDS_Shape & shape) const;
 
     void Analyse(Mesh& mesh,
                  const MeshingParameters& mparam) const override;
@@ -420,8 +427,8 @@ namespace netgen
     //bool FastProject (int surfi, Point<3> & ap, double& u, double& v) const;
   };
 
-  void Identify(const ListOfShapes & me, const ListOfShapes & you, string name, Identifications::ID_TYPE type, Transformation<3> trafo);
-  void Identify(const TopoDS_Shape & me, const TopoDS_Shape & you, string name, Identifications::ID_TYPE type, std::optional<std::variant<gp_Trsf, gp_GTrsf>> opt_trafo);
+  DLL_HEADER void Identify(const ListOfShapes & me, const ListOfShapes & you, string name, Identifications::ID_TYPE type, Transformation<3> trafo);
+  DLL_HEADER void Identify(const TopoDS_Shape & me, const TopoDS_Shape & you, string name, Identifications::ID_TYPE type, std::optional<std::variant<gp_Trsf, gp_GTrsf>> opt_trafo);
    
 
   void PrintContents (OCCGeometry * geom);
@@ -516,11 +523,13 @@ namespace netgen
                       if(from.IsSame(from_mapped) && to.IsSame(to_mapped))
                         continue;
   
-                      Transformation<3> trafo_mapped = ident.trafo;
+                      if(!ident.trafo) continue;
+                      Transformation<3> trafo_mapped = *ident.trafo;
+
                       if(trafo)
                       {
                           Transformation<3> trafo_temp;
-                          trafo_temp.Combine(ident.trafo, trafo_inv);
+                          trafo_temp.Combine(*ident.trafo, trafo_inv);
                           trafo_mapped.Combine(*trafo, trafo_temp);
                       }
   
