@@ -266,7 +266,7 @@ namespace netgen
 
     
       }
-    catch (bad_weak_ptr e)
+    catch (const bad_weak_ptr & e)
       {
         // cout << "don't have a mesh to visualize" << endl;
         VisualScene::DrawScene();      
@@ -292,11 +292,11 @@ namespace netgen
       center.Y() = vispar.centery;
       center.Z() = vispar.centerz;
     }
-    else if (selpoint >= 1 && zoomall==2)
+    else if (selpoint-IndexBASE<PointIndex>() >= 1 && zoomall==2)
       center = mesh->Point (selpoint);
     else if (marker && zoomall==2)
       center = *marker;
-    else if (vispar.centerpoint >= 1 && zoomall==2)
+    else if (vispar.centerpoint-IndexBASE<PointIndex>() >= 0 && zoomall==2)
       center = mesh->Point (vispar.centerpoint);
     else
       center = Center (pmin, pmax);
@@ -472,7 +472,7 @@ namespace netgen
 		const Point3d & p2 = mesh->Point(v.Elem(2));
 		const Point3d & p3 = mesh->Point(v.Elem(3));
 		Point3d p;
-		if (v.Elem(4) == 0)
+		if (v.Size() == 3)
                   {
 		    p = Center (p1, p2, p3);
                   }
@@ -515,20 +515,21 @@ namespace netgen
         if (vispar.drawelementnumbers)
 	  {
 	    NgArray<int> v;
-	    for (int i = 1; i <= mesh->GetNE(); i++)
+	    // for (int i = 1; i <= mesh->GetNE(); i++)
+            for (ElementIndex ei : Range(mesh->VolumeElements()))
 	      {
 		// const ELEMENTTYPE & eltype = mesh->ElementType(i);
 		NgArray<int> pnums;
 
 		Point3d p;
-		const Element & el = mesh->VolumeElement (i);
+		const Element & el = mesh->VolumeElement (ei);
 
 		if ( ! el.PNum(5)) //  eltype == TET )
                   {
 
 		    pnums.SetSize(4);
 		    for( int j = 0; j < pnums.Size(); j++)
-		      pnums[j] = mesh->VolumeElement(i).PNum(j+1);
+		      pnums[j] = mesh->VolumeElement(ei).PNum(j+1);
 
 
 		    const Point3d & p1 = mesh->Point(pnums[0]);
@@ -541,7 +542,7 @@ namespace netgen
                   {
 		    pnums.SetSize(5);
 		    for( int j = 0; j < pnums.Size(); j++)
-		      pnums[j] = mesh->VolumeElement(i).PNum(j+1);
+		      pnums[j] = mesh->VolumeElement(ei).PNum(j+1);
 
 
 		    const Point3d & p1 = mesh->Point(pnums[0]);
@@ -559,7 +560,7 @@ namespace netgen
                   {
 		    pnums.SetSize(6);
 		    for( int j = 0; j < pnums.Size(); j++)
-		      pnums[j] = mesh->VolumeElement(i).PNum(j+1);
+		      pnums[j] = mesh->VolumeElement(ei).PNum(j+1);
 
 		    const Point3d & p1 = mesh->Point(pnums[0]);
 		    const Point3d & p2 = mesh->Point(pnums[1]);
@@ -574,7 +575,7 @@ namespace netgen
                   {
 		    pnums.SetSize(8);
 		    for( int j = 0; j < pnums.Size(); j++)
-		      pnums[j] = mesh->VolumeElement(i).PNum(j+1);
+		      pnums[j] = mesh->VolumeElement(ei).PNum(j+1);
 
 		    const Point3d & p1 = mesh->Point(pnums[0]);
 		    const Point3d & p2 = mesh->Point(pnums[1]);
@@ -589,7 +590,7 @@ namespace netgen
                   }
 
 		glRasterPos3d (p.X(), p.Y(), p.Z());
-		snprintf (buf, size(buf),  "%d", i);
+		snprintf (buf, size(buf),  "%d", ei-IndexBASE(ei));
 		// glCallLists (strlen (buf), GL_UNSIGNED_BYTE, buf);
 		MyOpenGLText (buf);
 
@@ -617,14 +618,15 @@ namespace netgen
 	static float badelcol[] = { 1.0f, 0.0f, 1.0f, 1.0f };
 	glLineWidth (1.0f);
 
-	for (int i = 1; i <= mesh->GetNE(); i++)
+	//for (int i = 1; i <= mesh->GetNE(); i++)
+        for (ElementIndex ei : Range(mesh->VolumeElements()))
 	  {
-            if (mesh->VolumeElement(i).Flags().badel ||
-		mesh->VolumeElement(i).Flags().illegal ||
-		(i == vispar.drawelement))
+            if (mesh->VolumeElement(ei).Flags().badel ||
+		mesh->VolumeElement(ei).Flags().illegal ||
+		(ei-IndexBASE(ei) == vispar.drawelement))
 	      {
 		// copy to be thread-safe
-		Element el = mesh->VolumeElement (i);
+		Element el = mesh->VolumeElement (ei);
 		el.GetSurfaceTriangles (faces);
 
 		glMaterialfv(GL_FRONT, GL_AMBIENT_AND_DIFFUSE, badelcol);
@@ -674,9 +676,9 @@ namespace netgen
 	  }
 
 
-	for (int i = 1; i <= mesh->GetNE(); i++)
+        for (ElementIndex ei : Range(mesh->VolumeElements()))
 	  {
-            Element el = mesh->VolumeElement (i);
+            Element el = mesh->VolumeElement (ei);
             int hascp = 0;
             for (int j = 1; j <= el.GetNP(); j++)
 	      if (el.PNum(j) == vispar.centerpoint)
@@ -684,7 +686,7 @@ namespace netgen
 
             if (hascp)
 	      {
-		(*testout) << "draw el " << i << " : ";
+		(*testout) << "draw el " << ei << " : ";
 		for (int j = 1; j <= el.GetNP(); j++)
                   (*testout) << el.PNum(j) << " ";
 		(*testout) << endl;
@@ -851,37 +853,42 @@ namespace netgen
 	GLfloat identifiedcol[] = { 1, 0, 1, 1 };
 
 	glLineWidth (3);
-
-	//  for (i = 1; i <= mesh->GetNSeg(); i++)
+        glEnable (GL_COLOR_MATERIAL);
+        glDisable (GL_LIGHTING);
 
 	if (mesh -> HasIdentifications() )
 	  {
-            // if (mesh->GetIdentifications().HasIdentifiedPoints())
 	      {
-                INDEX_2_HASHTABLE<int> & idpts =
+                auto & idpts =
                   mesh->GetIdentifications().GetIdentifiedPoints();
-                
-		for (int i = 1; i <= idpts.GetNBags(); i++)
-                  for (int j = 1; j <= idpts.GetBagSize(i); j++)
-		    {
-		      INDEX_2 pts;
-		      int val;
-
-		      idpts.GetData (i, j, pts, val);
-		      const Point3d & p1 = mesh->Point(pts.I1());
-		      const Point3d & p2 = mesh->Point(pts.I2());
+                for (auto [hash, val] : idpts)
+                  {
+                    auto [hash_pts, hash_nr] = hash;
+                    auto [pi1, pi2] = hash_pts;
+                      // val = pts[2];   
+		      Point<3> p1 = mesh->Point(pi1);
+		      Point<3> p2 = mesh->Point(pi2);
+                      Point<3> c = Center(p1, p2);
+                      if (vispar.shrink < 1)
+                        {
+                          p1 = c + vispar.shrink * (p1 - c);
+                          p2 = c + vispar.shrink * (p2 - c);
+                        }
 
+                      glColor3fv (identifiedcol);
 		      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE,
 				    identifiedcol);
 
 		      glBegin (GL_LINES);
-		      glVertex3f (p1.X(), p1.Y(), p1.Z());
-		      glVertex3f (p2.X(), p2.Y(), p2.Z());
+		      glVertex3dv(p1);
+		      glVertex3dv(p2);
 		      glEnd();
 		    }
 	      }
 	  }
 
+        glDisable (GL_COLOR_MATERIAL);
+        glEnable (GL_LIGHTING);
 	glEndList ();
       }
 
@@ -894,7 +901,7 @@ namespace netgen
 
     vstimestamp = meshtimestamp;
       }
-    catch (bad_weak_ptr e)
+    catch (const bad_weak_ptr & e)
       {
         PrintMessage (3, "vsmesh::buildscene: don't have a mesh to visualize");
         VisualScene::BuildScene (zoomall);
